"""Unit tests for the parse_segv general hook."""

import unittest

from problem_report import ProblemReport
from tests.helper import import_module_from_file
from tests.paths import get_data_directory

parse_segv = import_module_from_file(
    get_data_directory() / "general-hooks" / "parse_segv.py"
)

# Default global registers, maps, and disassembly for testing
REGS = """\
eax            0xffffffff -1
ecx            0xbfc6af40   -1077498048
edx            0x1  1
ebx            0x26eff4 2551796
esp            0xbfc6af24   0xbfc6af24
ebp            0xbfc6af28   0xbfc6af28
esi            0x826bb60    136756064
edi            0x8083480    134755456
eip            0x808354e    0x808354e <main+14>
eflags         0x200286 [ PF SF IF ID ]
cs             0x73 115
ss             0x7b 123
ds             0x7b 123
es             0x7b 123
fs             0x4  4
gs             0x33 51
"""
REGS64 = """\
rax            0xffffffffffffffff   -1
rbx            0x26eff4 2551796
rcx            0xffffffffffffffff   -1
rdx            0xffffffffff600180   -10485376
rsi            0x0  0
rdi            0x7fffffffe3b0   140737488348080
rbp            0x0  0x0
rsp            0x0000bfc6af24   0x0000bfc6af24
r8             0x0  0
r9             0x0  0
r10            0x7fffffffe140   140737488347456
r11            0x246    582
r12            0x7fffffffe400   140737488348160
r13            0x7fffffffe468   140737488348264
r14            0x1  1
r15            0x7fffffffe460   140737488348256
rip            0x7ffff790be10   0x7ffff790be10 <nanosleep+16>
eflags         0x246    [ PF ZF IF ]
cs             0x33 51
ss             0x2b 43
ds             0x0  0
es             0x0  0
fs             0x0  0
gs             0x0  0
fctrl          0x37f    895
fstat          0x0  0
ftag           0xffff   65535
fiseg          0x0  0
fioff          0x40303a 4206650
foseg          0x0  0
fooff          0x0  0
fop            0x5d8    1496
mxcsr          0x1f80   [ IM DM ZM OM UM PM ]
"""
REGS64VIRT = """\
rax            0x0                 0
rbx            0x0                 0
rcx            0x555555557e08      93824992247304
rdx            0x2a                42
rsi            0x7fffffffedb8      140737488350648
rdi            0x2a                42
rbp            0x7fffffffec90      0x7fffffffec90
rsp            0x7fffffffec90      0x7fffffffec90
r8             0x7ffff7fa6680      140737353770624
r9             0x7ffff7fa7fe0      140737353777120
r10            0x7fffffffe9f0      140737488349680
r11            0x202               514
r12            0x1                 1
r13            0x7ffff7ffd000      140737354125312
r14            0x7fffffffedc8      140737488350664
r15            0x555555557e08      93824992247304
rip            0x55555555513f      0x55555555513f <f+22>
eflags         0x10246             [ PF ZF IF RF ]
cs             0x33                51
ss             0x2b                43
ds             0x0                 0
es             0x0                 0
fs             0x0                 0
gs             0x0                 0
pl3_ssp        <unavailable>
fs_base        0x7ffff7d90740      140737351583552
gs_base        0x0                 0
"""
MAPS = """\
00110000-0026c000 r-xp 00000000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026c000-0026d000 ---p 0015c000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026d000-0026f000 r--p 0015c000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026f000-00270000 rw-p 0015e000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
00270000-00273000 rw-p 00000000 00:00 0
002c1000-002e5000 r-xp 00000000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
002e5000-002e6000 r--p 00023000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
002e6000-002e7000 rw-p 00024000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
00318000-00334000 r-xp 00000000 08:06 977846     /lib/ld-2.9.so
00334000-00335000 r--p 0001b000 08:06 977846     /lib/ld-2.9.so
00335000-00336000 rw-p 0001c000 08:06 977846     /lib/ld-2.9.so
0056e000-005a1000 r-xp 00000000 08:06 65575      /lib/libncurses.so.5.7
005a1000-005a3000 r--p 00033000 08:06 65575      /lib/libncurses.so.5.7
005a3000-005a4000 rw-p 00035000 08:06 65575      /lib/libncurses.so.5.7
00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]
00bb6000-00bcb000 r-xp 00000000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcb000-00bcc000 r--p 00014000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcc000-00bcd000 rw-p 00015000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcd000-00bcf000 rw-p 00000000 00:00 0
00beb000-00bed000 r-xp 00000000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00bed000-00bee000 r--p 00001000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00bee000-00bef000 rw-p 00002000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00c56000-00c7a000 r-xp 00000000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00c7a000-00c7c000 r--p 00023000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00c7c000-00c7d000 rw-p 00025000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00dce000-00dfa000 r-xp 00000000 08:06 65612      /lib/libreadline.so.5.2
00dfa000-00dfb000 ---p 0002c000 08:06 65612      /lib/libreadline.so.5.2
00dfb000-00dfc000 r--p 0002c000 08:06 65612      /lib/libreadline.so.5.2
00dfc000-00dff000 rw-p 0002d000 08:06 65612      /lib/libreadline.so.5.2
00dff000-00e00000 rw-p 00000000 00:00 0
08048000-0831c000 r-xp 00000000 08:06 1140349    /usr/bin/gdb
0831c000-0831d000 r--p 002d3000 08:06 1140349    /usr/bin/gdb
0831d000-08325000 rw-p 002d4000 08:06 1140349    /usr/bin/gdb
08325000-0833f000 rw-p 00000000 00:00 0
b8077000-b807a000 rw-p 00000000 00:00 0
b8096000-b8098000 rw-p 00000000 00:00 0
bfc57000-bfc6c000 rw-p 00000000 00:00 0          [stack]
"""
DISASM = """\
0x08083540 <main+0>:    lea    0x4(%esp),%ecx
0x08083544 <main+4>:    and    $0xfffffff0,%esp
0x08083547 <main+7>:    pushl  -0x4(%ecx)
0x0808354a <main+10>:   push   %ebp
0x0808354b <main+11>:   mov    %esp,%ebp
0x0808354d <main+13>:   push   %ecx
0x0808354e <main+14>:   sub    $0x14,%esp
0x08083551 <main+17>:   mov    (%ecx),%eax
0x08083553 <main+19>:   mov    0x4(%ecx),%edx
0x08083556 <main+22>:   lea    -0x14(%ebp),%ecx
0x08083559 <main+25>:   movl   $0x0,-0xc(%ebp)
0x08083560 <main+32>:   movl   $0x826bc68,-0x8(%ebp)
0x08083567 <main+39>:   mov    %eax,-0x14(%ebp)
0x0808356a <main+42>:   mov    %edx,-0x10(%ebp)
0x0808356d <main+45>:   mov    %ecx,(%esp)
0x08083570 <main+48>:   call   0x8083580 <gdb_main>
0x08083575 <main+53>:   add    $0x14,%esp
0x08083578 <main+56>:   pop    %ecx
0x08083579 <main+57>:   pop    %ebp
0x0808357a <main+58>:   lea    -0x4(%ecx),%esp
0x0808357d <main+61>:   ret
"""
DISASM64 = """\
=> 0x55555555513f <f+22>:	mov    %edx,(%rax)
   0x555555555141 <f+24>:	mov    -0x14(%rbp),%eax
   0x555555555144 <f+27>:	add    $0x1,%eax
   0x555555555147 <f+30>:	pop    %rbp
   0x555555555148 <f+31>:	ret
   0x555555555149 <main>:	push   %rbp
   0x55555555514a <main+1>:	mov    %rsp,%rbp
   0x55555555514d <main+4>:	mov    $0x2a,%edi
   0x555555555152 <main+9>:	call   0x555555555129 <f>
   0x555555555157 <main+14>:	pop    %rbp
   0x555555555158 <main+15>:	ret
   0x555555555159:	add    %al,(%rax)
   0x55555555515b:	add    %cl,-0x7d(%rax)
   0x55555555515e <_fini+2>:	in     (%dx),%al
   0x55555555515f <_fini+3>:	or     %cl,-0x7d(%rax)
   0x555555555162 <_fini+6>:	(bad)
"""


class TestHookParseSegv(unittest.TestCase):
    """Test Segfault Parser."""

    def test_add_info_without_segv_reason(self) -> None:
        """Test add_info() with no SegvReason in result."""
        report = ProblemReport()
        report["Signal"] = "11"
        report["Architecture"] = "amd64"
        report["Disassembly"] = DISASM
        report["ProcMaps"] = MAPS
        report["Registers"] = REGS64

        parse_segv.add_info(report)

        self.assertEqual(
            report.get("SegvAnalysis"),
            "Segfault happened at: 0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
            "PC (0x08083540) ok\n"
            "insn (lea) does not access VMA\n"
            "SP (0xbfc6af24) ok\n"
            "Reason could not be automatically determined.",
        )
        self.assertNotIn("SegvReason", report)

    def test_invalid_00_registers(self):
        """Require valid registers."""
        regs = "a 0x10\nb !!!\n"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, "", "")
        try:
            segv = parse_segv.ParseSegv(regs, "", "")
        except ValueError as error:
            self.assertIn("invalid literal for int()", str(error))

        regs = "a 0x10"
        disasm = "0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.regs["a"], 0x10, segv)

        segv.regs = None
        self.assertRaises(ValueError, segv.parse_disassembly, "")

    def test_invalid_01_disassembly(self):
        # TODO: Split into separate test cases
        # pylint: disable=too-many-statements
        """Require valid disassembly."""
        regs = "a 0x10"

        disasm = ""
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "Dump ..."
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "Dump ...\nmonkey"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "monkey"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "0x1111111111: Cannot access memory at address 0x1111111111\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x1111111111, segv.pc)
        self.assertIsNone(segv.insn, segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x2111111111: \n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x2111111111, segv.pc)
        self.assertIsNone(segv.insn, segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x8069ff0 <fopen@plt+132220>: cmpb   $0x0,(%eax,%ebx,1)\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x8069FF0, segv.pc)
        self.assertEqual(segv.insn, "cmpb", segv.insn)
        self.assertEqual(segv.src, "$0x0", segv.src)
        self.assertEqual(segv.dest, "(%eax,%ebx,1)", segv.dest)

        disasm = "0xb765bb48 <_XSend+440>:  call   *0x40(%edi)\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB765BB48, segv.pc)
        self.assertEqual(segv.insn, "call", segv.insn)
        self.assertEqual(segv.src, "*0x40(%edi)", segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0xb7aae5a0:   call   0xb7a805af <_Unwind_Find_FDE@plt+111>\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB7AAE5A0, segv.pc)
        self.assertEqual(segv.insn, "call", segv.insn)
        self.assertEqual(segv.src, "0xb7a805af", segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x09083540:    mov    0x4(%esp),%es:%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x09083540, segv.pc)
        self.assertEqual(segv.insn, "mov", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%es:%ecx", segv.dest)

        disasm = "0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083540, segv.pc)
        self.assertEqual(segv.insn, "lea", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%ecx", segv.dest)

        disasm = (
            "0x404127 <exo_mount_hal_device_mount+167>:\n"
            "    repz cmpsb %es:(%rdi),%ds:(%rsi)\n"
        )
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x0404127, segv.pc)
        self.assertEqual(segv.insn, "repz cmpsb", segv.insn)
        self.assertEqual(segv.src, "%es:(%rdi)", segv.src)
        self.assertEqual(segv.dest, "%ds:(%rsi)", segv.dest)

        disasm = "0xb031765a <hufftab16+570>: add    0x3430433,%eax"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB031765A, segv.pc)
        self.assertEqual(segv.insn, "add", segv.insn)
        self.assertEqual(segv.src, "0x3430433", segv.src)
        self.assertEqual(segv.dest, "%eax", segv.dest)

        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083540, segv.pc)
        self.assertEqual(segv.insn, "lea", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%ecx", segv.dest)

        disasm = "0x08083550 <main+0>:    nop\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083550, segv.pc)
        self.assertEqual(segv.insn, "nop", segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        regs = "esp 0x444"
        disasm = "0x08083560 <main+0>:    push %ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083560, segv.pc)
        self.assertEqual(segv.insn, "push", segv.insn)
        self.assertEqual(segv.src, "%ecx", segv.src)
        self.assertEqual(segv.dest, "(%esp)", segv.dest)

        # GDB 7.1
        regs = "esp 0x444"
        disasm = "=> 0x08083560 <main+0>:    push %ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083560, segv.pc)
        self.assertEqual(segv.insn, "push", segv.insn)
        self.assertEqual(segv.src, "%ecx", segv.src)
        self.assertEqual(segv.dest, "(%esp)", segv.dest)

    def test_unavailable_register(self) -> None:
        """Test pl3_ssp register value being unavailable."""
        segv = parse_segv.ParseSegv(REGS64VIRT, DISASM64, "")
        self.assertNotIn("pl3_ssp", segv.regs)
        self.assertLessEqual(
            {"rax", "rbx", "rcx", "rdx", "rsi", "rdi", "rbp", "rsp"}, set(segv.regs)
        )
        self.assertLessEqual({f"r{i}" for i in range(8, 16)}, set(segv.regs))

    def test_ioport_operation(self):
        """I/O port violations."""
        regs = "rax            0x3  3"
        disasm = (
            "0x4087f1 <snd_pcm_hw_params_set_channels_near@plt+19345>:\n"
            "    out    %al,$0xb3\n"
        )
        maps = (
            "00400000-00412000 r-xp 00000000 08:04 10371157                   "
            "        /usr/sbin/pommed\n"
            "00611000-00614000 rw-p 00011000 08:04 10371157                   "
            "        /usr/sbin/pommed\n"
            "00614000-00635000 rw-p 00614000 00:00 0                          "
            "        [heap]\n"
        )
        segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertEqual(segv.pc, 0x4087F1, segv.pc)
        self.assertEqual(segv.insn, "out", segv.insn)
        self.assertEqual(segv.src, "%al", segv.src)
        self.assertEqual(segv.dest, "$0xb3", segv.dest)

        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertTrue("disallowed I/O port operation on port 3" in reason, reason)

    def test_invalid_02_maps(self):
        """Require valid maps."""
        regs = "a 0x10"
        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"

        maps = "asdlkfjaadf"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, maps)

        maps = (
            "005a3000-005a4000 rw-p 00035000 08:06 65575  "
            "    /lib/libncurses.so.5.7\n"
            "00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]\n"
            "00c67000-00c68000 r--p 00000000 00:00 0 "
        )
        segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertEqual(segv.maps[0]["start"], 0x005A3000, segv)
        self.assertEqual(segv.maps[0]["end"], 0x005A4000, segv)
        self.assertEqual(segv.maps[0]["perms"], "rw-p", segv)
        self.assertEqual(segv.maps[0]["name"], "/lib/libncurses.so.5.7", segv)

        self.assertEqual(segv.maps[1]["start"], 0x00B67000, segv)
        self.assertEqual(segv.maps[1]["end"], 0x00B68000, segv)
        self.assertEqual(segv.maps[1]["perms"], "r-xp", segv)
        self.assertEqual(segv.maps[1]["name"], "[vdso]", segv)

        self.assertEqual(segv.maps[2]["start"], 0x00C67000, segv)
        self.assertEqual(segv.maps[2]["end"], 0x00C68000, segv)
        self.assertEqual(segv.maps[2]["perms"], "r--p", segv)
        self.assertIsNone(segv.maps[2]["name"], segv)

    def test_debug(self):
        """Debug mode works."""
        regs = "a 0x10"
        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        maps = (
            "005a3000-005a4000 rw-p 00035000 08:06 65575  "
            "    /lib/libncurses.so.5.7\n"
            "00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]\n"
            "00c67000-00c68000 r--p 00000000 00:00 0 "
        )
        with self.assertLogs(level="DEBUG"):
            segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertTrue(segv is not None, segv)

    def test_register_values(self):
        """Sub-register parsing."""
        disasm = """0x08083540 <main+0>:    mov    $1,%ecx"""
        segv = parse_segv.ParseSegv(REGS64, disasm, "")

        val = segv.register_value("%rdx")
        self.assertEqual(val, 0xFFFFFFFFFF600180, hex(val))
        val = segv.register_value("%edx")
        self.assertEqual(val, 0xFF600180, hex(val))
        val = segv.register_value("%dx")
        self.assertEqual(val, 0x0180, hex(val))
        val = segv.register_value("%dl")
        self.assertEqual(val, 0x80, hex(val))

    def test_segv_unknown(self):
        """Handle unknown segfaults."""
        disasm = """0x08083540 <main+0>:    mov    $1,%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)

        # Verify calculations
        self.assertEqual(segv.calculate_arg("(%ecx)"), 0xBFC6AF40, segv.regs["ecx"])
        self.assertEqual(segv.calculate_arg("0x10(%ecx)"), 0xBFC6AF50, segv.regs["ecx"])
        self.assertEqual(
            segv.calculate_arg("-0x20(%ecx)"), 0xBFC6AF20, segv.regs["ecx"]
        )
        self.assertEqual(segv.calculate_arg("%fs:(%ecx)"), 0xBFC6AF44, segv.regs["ecx"])
        self.assertEqual(segv.calculate_arg("0x3404403"), 0x3404403, "0x3404403")
        self.assertEqual(segv.calculate_arg("*0x40(%edi)"), 0x80834C0, segv.regs["edi"])
        self.assertEqual(
            segv.calculate_arg("(%edx,%ebx,1)"), 0x26EFF5, segv.regs["ebx"]
        )
        self.assertEqual(
            segv.calculate_arg("(%eax,%ebx,1)"), 0x26EFF3, segv.regs["ebx"]
        )
        self.assertEqual(
            segv.calculate_arg("0x10(,%ebx,1)"), 0x26F004, segv.regs["ebx"]
        )

        # Again, but 64bit
        disasm = """0x08083540 <main+0>:    mov    $1,%rcx"""
        segv = parse_segv.ParseSegv(REGS64, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)

        self.assertEqual(
            segv.calculate_arg("(%rax,%rbx,1)"), 0x26EFF3, segv.regs["rbx"]
        )

    def test_segv_pc_missing(self):
        """Handle PC in missing VMA."""
        disasm = """0x00083540 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00083540) not located in a known VMA region", details)
        self.assertIn("executing unknown VMA", reason)

        disasm = """0x00083544:"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00083544) not located in a known VMA region", details)
        self.assertIn("executing unknown VMA", reason)

    def test_segv_pc_null(self):
        """Handle PC in NULL VMA."""
        disasm = """0x00000540 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00000540) not located in a known VMA region", details)
        self.assertIn("executing NULL VMA", reason)

    def test_segv_pc_nx_writable(self):
        """Handle PC in writable NX VMA."""
        disasm = """0x005a3000 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x005a3000) in non-executable VMA region:", details)
        self.assertIn("executing writable VMA /lib/libncurses.so.5.7", reason)

    def test_segv_pc_nx_unwritable(self):
        """Handle PC in non-writable NX VMA."""
        disasm = """0x00dfb000 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00dfb000) in non-executable VMA region:", details)
        self.assertIn("executing non-writable VMA /lib/libreadline.so.5.2", reason)

    def test_segv_src_missing(self):
        """Handle source in missing VMA."""
        reg = f"{REGS}ecx            0x0006af24   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        # Valid crash
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x0006af20) not located in a known VMA region',
            details,
        )
        self.assertIn("reading unknown VMA", reason)

        # Valid crash
        disasm = "0x08083547 <main+7>:    callq  *%ecx"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "*%ecx" (0x0006af24) not located in a known VMA region', details
        )
        self.assertIn("reading unknown VMA", reason)

    def test_segv_src_null(self):
        """Handle source in NULL VMA."""
        reg = f"{REGS}ecx            0x00000024   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x00000020) not located in a known VMA region',
            details,
        )
        self.assertIn("reading NULL VMA", reason)

    def test_segv_src_not_readable(self):
        """Handle source not in readable VMA."""
        reg = f"{REGS}ecx            0x0026c080   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x0026c07c) in non-readable VMA region:', details
        )
        self.assertIn("reading VMA /lib/tls/i686/cmov/libc-2.9.so", reason)
        self.assertNotIn("Stack memory exhausted", details)
        self.assertNotIn("Stack pointer not within stack segment", details)

    def test_segv_dest_missing(self):
        """Handle destination in missing VMA."""
        reg = f"{REGS}esp            0x0006af24   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x0006af24) not located in a known VMA region',
            details,
        )
        self.assertIn("writing unknown VMA", reason)

    def test_segv_dest_null(self):
        """Handle destination in NULL VMA."""
        reg = f"{REGS}esp            0x00000024   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x00000024) not located in a known VMA region',
            details,
        )
        self.assertIn("writing NULL VMA", reason)

    def test_segv_dest_not_writable(self):
        """Handle destination not in writable VMA."""
        reg = f"{REGS}esp            0x08048080   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x08048080) in non-writable VMA region:', details
        )
        self.assertIn("writing VMA /usr/bin/gdb", reason)

    def test_segv_crackful_disasm(self):
        """Reject insane disassemblies."""
        disasm = "0x08083547 <main+7>:    pushl  -0x4(blah)"
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        self.assertRaises(ValueError, segv.report)

        disasm = "0x08083547 <main+7>:    pushl  -04(%ecx)"
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        self.assertRaises(ValueError, segv.report)

    def test_segv_stack_failure(self):
        """Handle walking off the stack."""
        # Triggered via "push"
        reg = f"{REGS}esp            0xbfc56ff0   0xbfc56ff0"
        disasm = "0x08083547 <main+7>:    push  %eax"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0xbfc56ff0) not located'
            " in a known VMA region (needed writable region)!",
            details,
        )

        # Triggered via "call"
        reg = f"{REGS}esp            0xbfc56fff   0xbfc56fff"
        disasm = "0x08083547 <main+7>:    callq  0x08083540"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0xbfc56fff) not located'
            " in a known VMA region (needed writable region)!",
            details,
        )
        self.assertIn("Stack memory exhausted", details)

        # Triggered via unknown reason
        reg = f"{REGS}esp            0xdfc56000   0xdfc56000"
        disasm = """0x08083540 <main+0>:    mov    $1,%rcx"""
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            "SP (0xdfc56000) not located in a known VMA region"
            " (needed readable region)!",
            details,
        )
        self.assertIn("Stack pointer not within stack segment", details)

    def test_segv_stack_kernel_segfault(self):
        """Handle unknown segfaults in kernel."""
        # Crash in valid code path
        disasm = """0x0056e010: ret"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)
        self.assertIn("Reason could not be automatically determined.", details)
        self.assertNotIn("(Unhandled exception in kernel code?)", details)

        # Crash from kernel code path
        disasm = """0x00b67422 <__kernel_vsyscall+2>: ret"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)
        self.assertIn(
            "Reason could not be automatically determined."
            " (Unhandled exception in kernel code?)",
            details,
        )



"""Integration tests for the apport.hookutils module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import locale
import os
import re
import shutil
import subprocess
import sys
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apport.hookutils
import apport.report
import problem_report
from tests.helper import skip_if_command_is_missing


class T(unittest.TestCase):
    """Integration tests for the apport.hookutils module."""

    def setUp(self):
        self.workdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.workdir)

    @skip_if_command_is_missing("/usr/bin/as")
    def test_module_license_evaluation(self):
        """Module licenses can be validated correctly."""
        # pylint: disable=protected-access

        def _build_ko(license_name):
            ko_filename = os.path.join(self.workdir, f"{license_name}.ko")
            with tempfile.NamedTemporaryFile(
                mode="w+", prefix=f"{license_name}-", suffix=".S"
            ) as asm:
                asm.write(f'.section .modinfo\n.string "license={license_name}"\n')
                asm.flush()
                subprocess.check_call(["/usr/bin/as", asm.name, "-o", ko_filename])
            return ko_filename

        good_ko = _build_ko("GPL")
        bad_ko = _build_ko("BAD")

        # test:
        #  - unfindable module
        #  - fake GPL module
        #  - fake BAD module

        # direct license check
        self.assertEqual(
            apport.hookutils._get_module_license("does-not-exist"), "invalid"
        )
        self.assertIn("GPL", apport.hookutils._get_module_license(good_ko))
        self.assertIn("BAD", apport.hookutils._get_module_license(bad_ko))

        # check via nonfree_kernel_modules logic
        with tempfile.NamedTemporaryFile(mode="w+") as temp:
            temp.write(f"isofs\ndoes-not-exist\n{good_ko}\n{bad_ko}\n")
            temp.flush()
            nonfree = apport.hookutils.nonfree_kernel_modules(temp.name)
        self.assertIn("does-not-exist", nonfree)
        self.assertNotIn(good_ko, nonfree)
        self.assertIn(bad_ko, nonfree)

    @skip_if_command_is_missing("/sbin/modinfo")
    def test_real_module_license_evaluation(self):
        """Module licenses can be validated correctly for real module."""
        # pylint: disable=protected-access
        isofs_license = apport.hookutils._get_module_license("isofs")
        if isofs_license == "invalid":
            self.skipTest("kernel module 'isofs' not available")

        self.assertIn("GPL", isofs_license)

        with tempfile.NamedTemporaryFile() as temp:
            temp.write(b"isofs\n")
            temp.flush()
            nonfree = apport.hookutils.nonfree_kernel_modules(temp.name)
        self.assertNotIn("isofs", nonfree)

    def test_attach_file(self):
        """attach_file()"""
        with open("/etc/passwd", encoding="utf-8") as f:
            passwd_contents = f.read().strip()
        with open("/etc/issue", encoding="utf-8") as f:
            issue_contents = f.read().strip()

        # default key name
        report = {}
        apport.hookutils.attach_file(report, "/etc/passwd")
        self.assertEqual(list(report), [".etc.passwd"])
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        # custom key name
        report = {}
        apport.hookutils.attach_file(report, "/etc/passwd", "Passwd")
        self.assertEqual(list(report), ["Passwd"])
        self.assertEqual(report["Passwd"], passwd_contents)

        # nonexisting file
        report = {}
        apport.hookutils.attach_file(report, "/nonexisting")
        self.assertEqual(list(report), [".nonexisting"])
        self.assertTrue(report[".nonexisting"].startswith("Error: "))

        # symlink
        link = os.path.join(self.workdir, "symlink")
        os.symlink("/etc/passwd", link)
        report = {}
        apport.hookutils.attach_file(report, link, "Symlink")
        self.assertEqual(list(report), ["Symlink"])
        self.assertTrue(report["Symlink"].startswith("Error: "))

        # directory symlink
        link = os.path.join(self.workdir, "dirsymlink")
        os.symlink("/etc", link)
        report = {}
        apport.hookutils.attach_file(report, os.path.join(link, "passwd"), "DirSymlink")
        self.assertEqual(list(report), ["DirSymlink"])
        self.assertTrue(report["DirSymlink"].startswith("Error: "))

        # directory traversal
        report = {}
        apport.hookutils.attach_file(report, "/etc/../etc/passwd", "Traversal")
        self.assertEqual(list(report), ["Traversal"])
        self.assertTrue(report["Traversal"].startswith("Error: "))

        # existing key
        report = {}
        apport.hookutils.attach_file(report, "/etc/passwd")
        apport.hookutils.attach_file(report, "/etc/passwd")
        self.assertEqual(list(report), [".etc.passwd"])
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        apport.hookutils.attach_file(
            report, "/etc/issue", ".etc.passwd", overwrite=False
        )
        self.assertEqual(sorted(report.keys()), [".etc.passwd", ".etc.passwd_"])
        self.assertEqual(report[".etc.passwd"], passwd_contents)
        self.assertEqual(report[".etc.passwd_"], issue_contents)

    def test_attach_file_binary(self):
        """attach_file() for binary files"""
        myfile = os.path.join(self.workdir, "data")
        with open(myfile, "wb") as f:
            f.write(b"a\xc3\xb6b\xffx")

        report = {}
        apport.hookutils.attach_file(report, myfile, key="data")
        self.assertEqual(report["data"], b"a\xc3\xb6b\xffx")

        apport.hookutils.attach_file(report, myfile, key="data", force_unicode=True)
        self.assertEqual(report["data"], b"a\xc3\xb6b\xef\xbf\xbdx".decode("UTF-8"))

    def test_attach_file_if_exists(self):
        """attach_file_if_exists()"""
        with open("/etc/passwd", encoding="utf-8") as f:
            passwd_contents = f.read().strip()

        # default key name
        report = {}
        apport.hookutils.attach_file_if_exists(report, "/etc/passwd")
        self.assertEqual(list(report), [".etc.passwd"])
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        # custom key name
        report = {}
        apport.hookutils.attach_file_if_exists(report, "/etc/passwd", "Passwd")
        self.assertEqual(list(report), ["Passwd"])
        self.assertEqual(report["Passwd"], passwd_contents)

        # symlink
        link = os.path.join(self.workdir, "symlink")
        os.symlink("/etc/passwd", link)
        report = {}
        apport.hookutils.attach_file_if_exists(report, link, "Symlink")
        self.assertEqual(list(report), ["Symlink"])
        self.assertTrue(report["Symlink"].startswith("Error: "))

        # nonexisting file
        report = {}
        apport.hookutils.attach_file_if_exists(report, "/nonexisting")
        self.assertEqual(list(report), [])

        # directory traversal
        report = {}
        apport.hookutils.attach_file_if_exists(report, "/etc/../etc/passwd")
        self.assertEqual(list(report), [])

    def test_recent_syslog(self):
        """recent_syslog"""
        self.assertEqual(
            apport.hookutils.recent_syslog(re.compile("."), path="/nonexisting"), ""
        )
        self.assertEqual(
            apport.hookutils.recent_syslog(re.compile("ThisCantPossiblyHitAnything")),
            "",
        )
        if os.path.exists("/run/systemd/system") or os.access(
            "/var/log/syslog", os.R_OK
        ):
            self.assertNotEqual(len(apport.hookutils.recent_syslog(re.compile("."))), 0)

    @unittest.mock.patch(
        "apport.hookutils._root_command_prefix", MagicMock(return_value=[])
    )
    def test_attach_mac_events(self):
        # TODO: Split into separate test cases
        # pylint: disable=too-many-statements
        """attach_mac_events()"""
        denied_log = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/cupsd" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        denied_hex = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile=2F7573722F7362696E2F6375707364 pid=1361 comm="cupsd"'
            ' pid=1361 comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        report = problem_report.ProblemReport()
        apport.hookutils.attach_mac_events(report)
        self.assertIn("KernLog", report)

        # No AppArmor messages
        report = problem_report.ProblemReport()
        report["KernLog"] = (
            "[    2.997534] i915 0000:00:02.0:"
            " power state changed by ACPI to D0\n"
            "[    2.997541] i915 0000:00:02.0:"
            " PCI INT A -> GSI 16 (level, low)\n"
            "[    2.997544] i915 0000:00:02.0: setting latency timer to 64\n"
            "[    3.061584] i915 0000:00:02.0: irq 42 for MSI/MSI-X\n"
        )

        apport.hookutils.attach_mac_events(report)
        self.assertNotIn("Tags", report)

        # AppArmor message, but not a denial
        report = problem_report.ProblemReport()
        report["KernLog"] = (
            "[   32.420248] type=1400 audit(1344562672.449:2):"
            ' apparmor="STATUS" operation="profile_load" name="/sbin/dhclient"'
            ' pid=894 comm="apparmor_parser"\n'
        )

        apport.hookutils.attach_mac_events(report)
        self.assertNotIn("Tags", report)

        # AppArmor denial, empty tags, no profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor hex-encoded denial, no profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_hex

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial in AuditLog
        report = problem_report.ProblemReport()
        report["AuditLog"] = denied_log

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial, pre-existing tags, no profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log
        report["Tags"] = "bogustag"

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor bogustag")

        # AppArmor denial, single profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial, regex profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cups.*")
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial, subset profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cup")
        self.assertNotIn("Tags", report)

        # AppArmor hex-encoded denial, single profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_hex

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial, single different profile specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/nonexistent")
        self.assertNotIn("Tags", report)

        # AppArmor denial, multiple profiles specified
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log
        profiles = ["/usr/bin/nonexistent", "/usr/sbin/cupsd"]

        apport.hookutils.attach_mac_events(report, profiles)
        self.assertEqual(report["Tags"], "apparmor")

        # AppArmor denial, multiple different profiles
        report = problem_report.ProblemReport()
        report["KernLog"] = denied_log
        profiles = ["/usr/bin/nonexistent", "/usr/sbin/anotherone"]

        apport.hookutils.attach_mac_events(report, profiles)
        self.assertNotIn("Tags", report)

        # Multiple AppArmor denials, second match
        report = problem_report.ProblemReport()
        report["KernLog"] = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/blah" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/cupsd" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")

    def test_recent_syslog_overflow(self):
        """recent_syslog on a huge file"""
        log = os.path.join(self.workdir, "syslog")
        with open(log, "w", encoding="utf-8") as f:
            lines = 1000000
            while lines >= 0:
                f.write("Apr 20 11:30:00 komputer kernel: bogus message\n")
                lines -= 1

        mem_before = self._get_mem_usage()
        data = apport.hookutils.recent_syslog(re.compile("kernel"), path=log)
        mem_after = self._get_mem_usage()
        delta_kb = mem_after - mem_before
        sys.stderr.write(f"[Δ {delta_kb} kB] ")
        self.assertLess(delta_kb, 5000)

        self.assertRegex(data, "^Apr 20 11:30:00 komputer kernel: bogus message\n")
        self.assertGreater(len(data), 100000)
        self.assertLess(len(data), 1000000)

    @unittest.skipIf(
        apport.hookutils.apport.hookutils.in_session_of_problem(apport.report.Report())
        is None,
        "no logind session",
    )
    def test_in_session_of_problem(self):
        """in_session_of_problem()"""
        report = apport.report.Report(date="Sat Jan  1 12:00:00 2011")
        self.assertFalse(apport.hookutils.in_session_of_problem(report))

        report = apport.report.Report(date="Mon Oct 10 21:06:03 2009")
        self.assertFalse(apport.hookutils.in_session_of_problem(report))

        report = apport.report.Report()
        self.assertTrue(apport.hookutils.in_session_of_problem(report))

        self.assertIsNone(apport.hookutils.in_session_of_problem({}))

        orig_ctime = locale.getlocale(locale.LC_TIME)
        try:
            locale.setlocale(locale.LC_TIME, "C")

            report = apport.report.Report(date="Sat Jan  1 12:00:00 2011")
            self.assertFalse(apport.hookutils.in_session_of_problem(report))

            report = apport.report.Report(date="Mon Oct 10 21:06:03 2009")
            self.assertFalse(apport.hookutils.in_session_of_problem(report))

            report = apport.report.Report(date="Tue Jan  1 12:00:00 2038")
            self.assertTrue(apport.hookutils.in_session_of_problem(report))
        finally:
            locale.setlocale(locale.LC_TIME, orig_ctime)

    def test_xsession_errors(self):
        """xsession_errors()"""
        with open(
            os.path.join(self.workdir, ".xsession-errors"), "w", encoding="UTF-8"
        ) as f:
            f.write(
                """\
Loading profile from /etc/profile
gnome-session[1948]: WARNING: standard glib warning
EggSMClient-CRITICAL **: egg_sm_client_set_mode: standard glib assertion
24/02/2012 11:14:46 Sending credentials s3kr1t

** WARNING **: nonstandard warning

WARN  2012-02-24 11:23:47 unity <unknown>:0 some unicode ♥ ♪

GNOME_KEYRING_CONTROL=/tmp/keyring-u7hrD6

(gnome-settings-daemon:5115): Gdk-WARNING **: The program\
 'gnome-settings-daemon' received an X Window System error.
This probably reflects a bug in the program.
The error was 'BadMatch (invalid parameter attributes)'.
  (Details: serial 723 error_code 8 request_code 143 minor_code 22)
  (Note to programmers: normally, X errors are reported asynchronously;
   that is, you will receive the error a while after causing it.
   To debug your program, run it with the --sync command line
   option to change this behavior. You can then get a meaningful
   backtrace from your debugger if you break on the gdk_x_error() function.)"

GdkPixbuf-CRITICAL **: gdk_pixbuf_scale_simple: another standard glib assertion
"""
            )
        orig_home = os.environ.get("HOME")
        try:
            os.environ["HOME"] = self.workdir

            # explicit pattern
            pattern = re.compile("notfound")
            self.assertEqual(apport.hookutils.xsession_errors(pattern), "")

            pattern = re.compile(r"^\w+-CRITICAL")
            res = apport.hookutils.xsession_errors(pattern).splitlines()
            self.assertEqual(len(res), 2)
            self.assertTrue(res[0].startswith("EggSMClient-CRITICAL"))
            self.assertTrue(res[1].startswith("GdkPixbuf-CRITICAL"))

            # default pattern includes glib assertions and X Errors
            res = apport.hookutils.xsession_errors()
            self.assertNotIn("nonstandard warning", res)
            self.assertNotIn("keyring", res)
            self.assertNotIn("credentials", res)
            self.assertIn("WARNING: standard glib warning", res)
            self.assertIn("GdkPixbuf-CRITICAL", res)
            self.assertIn("'gnome-settings-daemon' received an X Window", res)
            self.assertIn("BadMatch", res)
            self.assertIn("serial 723", res)

        finally:
            if orig_home is not None:
                os.environ["HOME"] = orig_home
            else:
                os.unsetenv("HOME")

    @staticmethod
    @unittest.mock.patch(
        "apport.hookutils._root_command_prefix", MagicMock(return_value=[])
    )
    def test_no_crashes():
        """Functions do not crash (very shallow)."""
        report = {}
        apport.hookutils.attach_hardware(report)
        apport.hookutils.attach_alsa(report)
        apport.hookutils.attach_network(report)
        apport.hookutils.attach_wifi(report)
        apport.hookutils.attach_printing(report)
        apport.hookutils.attach_conffiles(report, "bash")
        apport.hookutils.attach_conffiles(report, "apport")
        apport.hookutils.attach_conffiles(report, "nonexisting")
        apport.hookutils.attach_default_grub(report)

    def test_command_output(self):
        """Test apport.hookutils.command_output."""
        orig_lcm = os.environ.get("LC_MESSAGES")
        os.environ["LC_MESSAGES"] = "en_US.UTF-8"
        try:
            # default mode: disable translations
            out = apport.hookutils.command_output(["env"])
            self.assertIn("LC_MESSAGES=C", out)

            # keep locale
            out = apport.hookutils.command_output(["env"], keep_locale=True)
            self.assertNotIn("LC_MESSAGES=C", out)
        finally:
            if orig_lcm is not None:
                os.environ["LC_MESSAGES"] = orig_lcm
            else:
                del os.environ["LC_MESSAGES"]

        # nonexisting binary
        out = apport.hookutils.command_output(["/non existing"])
        self.assertTrue(out.startswith("Error: [Errno 2]"))

        # stdin
        out = apport.hookutils.command_output(["cat"], input=b"hello")
        self.assertEqual(out, "hello")

    @staticmethod
    def _get_mem_usage():
        """Get current memory usage in kB."""
        with open("/proc/self/status", encoding="utf-8") as f:
            for line in f:
                if line.startswith("VmSize:"):
                    return int(line.split()[1])

        raise SystemError("did not find VmSize: in /proc/self/status")



# Copyright (C) 2023 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for data/general-hooks/wayland_session.py."""

import unittest

import problem_report
from tests.helper import import_module_from_file
from tests.paths import get_data_directory

wayland_session = import_module_from_file(
    get_data_directory() / "general-hooks" / "wayland_session.py"
)


class TestGeneralHookWaylandSession(unittest.TestCase):
    """Unit tests for data/general-hooks/wayland_session.py."""

    @unittest.mock.patch.dict("os.environ", {"WAYLAND_DISPLAY": "wayland-0"})
    def test_is_wayland_session(self) -> None:
        """Test add_info() for a Wayland session."""
        report = problem_report.ProblemReport()
        wayland_session.add_info(report, None)
        self.assertEqual(set(report.keys()), {"Date", "ProblemType", "Tags"})
        self.assertEqual(report["Tags"], "wayland-session")

    @unittest.mock.patch.dict("os.environ", {}, clear=True)
    def test_is_no_wayland_session(self) -> None:
        """Test add_info() for a session that isn't using Wayland."""
        report = problem_report.ProblemReport()
        wayland_session.add_info(report, None)
        self.assertEqual(set(report.keys()), {"Date", "ProblemType"})



"""Integration tests for the apport.fileutils module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import glob
import os
import pwd
import re
import shutil
import sys
import tempfile
import time
import unittest
import unittest.mock
from collections.abc import Iterator

import apport.fileutils
import apport.packaging
import problem_report


class T(unittest.TestCase):
    # pylint: disable=protected-access
    """Integration tests for the apport.fileutils module."""

    def setUp(self):
        self.orig_core_dir = apport.fileutils.core_dir
        apport.fileutils.core_dir = tempfile.mkdtemp()
        self.orig_report_dir = apport.fileutils.report_dir
        apport.fileutils.report_dir = tempfile.mkdtemp()
        self.orig_config_file = apport.fileutils._config_file

    def tearDown(self):
        shutil.rmtree(apport.fileutils.core_dir)
        apport.fileutils.core_dir = self.orig_core_dir
        shutil.rmtree(apport.fileutils.report_dir)
        apport.fileutils.report_dir = self.orig_report_dir
        self.orig_report_dir = None
        apport.fileutils._config_file = self.orig_config_file

    @staticmethod
    def _create_reports(create_inaccessible=False):
        """Create some test reports."""
        r1 = os.path.join(apport.fileutils.report_dir, "rep1.crash")
        r2 = os.path.join(apport.fileutils.report_dir, "rep2.crash")

        with open(r1, "w", encoding="utf-8") as fd:
            fd.write("report 1")
        with open(r2, "w", encoding="utf-8") as fd:
            fd.write("report 2")
        os.chmod(r1, 0o600)
        os.chmod(r2, 0o600)
        if create_inaccessible:
            ri = os.path.join(apport.fileutils.report_dir, "inaccessible.crash")
            with open(ri, "w", encoding="utf-8") as fd:
                fd.write("inaccessible")
            os.chmod(ri, 0)
            return [r1, r2, ri]
        return [r1, r2]

    @staticmethod
    def _packages_with_desktop_files() -> Iterator[tuple[str, int, int]]:
        desktop_path_re = re.compile("^/usr/share/applications/[^/]+.desktop$")
        for path in sorted(glob.glob("/usr/share/applications/*.desktop")):
            pkg = apport.packaging.get_file_package(path)
            if pkg is None:
                continue

            display_num = 0
            no_display_num = 0
            for desktop_file in apport.packaging.get_files(pkg):
                if not desktop_path_re.match(desktop_file):
                    continue
                with open(desktop_file, "rb") as desktop_file:
                    if b"NoDisplay=true" in desktop_file.read():
                        no_display_num += 1
                    else:
                        display_num += 1

            yield (pkg, display_num, no_display_num)

    def test_find_package_desktopfile_none(self) -> None:
        """find_package_desktopfile() for a package without any .desktop file"""
        nodesktop = "bash"
        assert (
            len(
                [
                    f
                    for f in apport.packaging.get_files(nodesktop)
                    if f.endswith(".desktop")
                ]
            )
            == 0
        )
        self.assertIsNone(
            apport.fileutils.find_package_desktopfile(nodesktop),
            f"no-desktop package {nodesktop}",
        )

    def test_find_package_desktopfile_one(self) -> None:
        """find_package_desktopfile() for a package with exactly one .desktop file"""
        for pkg, display_num, _ in self._packages_with_desktop_files():
            if display_num == 1:
                onedesktop = pkg
                break
        else:
            self.skipTest("no package with exactly one .desktop file found")

        d = apport.fileutils.find_package_desktopfile(onedesktop)
        self.assertIsNotNone(d, f"one-desktop package {onedesktop}")
        self.assertTrue(os.path.exists(d))
        self.assertTrue(d.endswith(".desktop"))

    def test_find_package_desktopfile_multiple(self) -> None:
        """find_package_desktopfile() for a package without multiple .desktop files"""
        for pkg, display_num, _ in self._packages_with_desktop_files():
            if display_num > 1:
                multidesktop = pkg
                break
        else:
            self.skipTest("no package with multiple .desktop files found")

        self.assertIsNone(
            apport.fileutils.find_package_desktopfile(multidesktop),
            f"multi-desktop package {multidesktop}",
        )

    def test_find_package_desktopfile_no_display(self) -> None:
        """find_package_desktopfile() for a package with a NoDisplay .desktop file"""
        nodisplay = None
        for pkg, display_num, no_display_num in self._packages_with_desktop_files():
            if display_num == 0 and no_display_num == 1:
                nodisplay = pkg
                break
        # expect a package with a NoDisplay .desktop file, e. g. python3.XX
        assert nodisplay is not None

        self.assertIsNone(
            apport.fileutils.find_package_desktopfile(nodisplay),
            f"NoDisplay package {nodisplay}",
        )

    def test_find_file_package(self):
        """find_file_package()"""
        self.assertEqual(apport.fileutils.find_file_package("/bin/bash"), "bash")
        self.assertEqual(apport.fileutils.find_file_package("/bin/cat"), "coreutils")
        self.assertIsNone(apport.fileutils.find_file_package("/nonexisting"))

    def test_seen(self):
        """get_new_reports() and seen_report()"""
        self.assertEqual(apport.fileutils.get_new_reports(), [])
        if os.getuid() == 0:
            tr = self._create_reports(True)
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
        self.assertEqual(set(apport.fileutils.get_new_reports()), set(tr))

        # now mark them as seen and check again
        nr = set(tr)
        for r in tr:
            self.assertEqual(apport.fileutils.seen_report(r), False)
            nr.remove(r)
            apport.fileutils.mark_report_seen(r)
            self.assertEqual(apport.fileutils.seen_report(r), True)
            self.assertEqual(set(apport.fileutils.get_new_reports()), nr)

    def test_mark_hanging_process(self):
        """mark_hanging_process()"""
        pr = problem_report.ProblemReport()
        pr["ExecutablePath"] = "/bin/bash"
        apport.fileutils.mark_hanging_process(pr, "1")
        uid = str(os.getuid())
        base = f"_bin_bash.{uid}.1.hanging"
        expected = os.path.join(apport.fileutils.report_dir, base)
        self.assertTrue(os.path.exists(expected))

    def test_mark_report_upload(self):
        """mark_report_upload()"""
        report = os.path.join(apport.fileutils.report_dir, "report.crash")
        apport.fileutils.mark_report_upload(report)
        expected = os.path.join(apport.fileutils.report_dir, "report.upload")
        self.assertTrue(os.path.exists(expected))

    def test_mark_2nd_report_upload(self):
        """mark_report_upload() for a previously uploaded report"""
        upload = os.path.join(apport.fileutils.report_dir, "report.upload")
        with open(upload, "w", encoding="utf-8"):
            pass
        uploaded = os.path.join(apport.fileutils.report_dir, "report.uploaded")
        with open(uploaded, "w", encoding="utf-8"):
            pass
        time.sleep(1)
        report = os.path.join(apport.fileutils.report_dir, "report.crash")
        with open(report, "w", encoding="utf-8"):
            pass
        time.sleep(1)
        apport.fileutils.mark_report_upload(report)
        upload_st = os.stat(upload)
        report_st = os.stat(report)
        self.assertTrue(upload_st.st_mtime > report_st.st_mtime)

    def test_get_all_reports(self):
        """get_all_reports()"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        if os.getuid() == 0:
            tr = self._create_reports(True)
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
        self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))

        # now mark them as seen and check again
        for r in tr:
            apport.fileutils.mark_report_seen(r)

        self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))

    def test_get_system_reports(self):
        """get_all_system_reports() and get_new_system_reports()"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])
        if os.getuid() <= apport.fileutils.get_sys_uid_max():
            tr = self._create_reports(True)
            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set(tr))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set(tr))

            # now mark them as seen and check again
            for r in tr:
                apport.fileutils.mark_report_seen(r)

            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set(tr))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set([]))
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set([]))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set([]))

    @unittest.mock.patch.object(os, "stat")
    @unittest.mock.patch.object(pwd, "getpwuid")
    def test_get_system_reports_guest(self, getpwuid_mock, stat_mock):
        """get_all_system_reports() filters out reports from guest user"""
        self._create_reports()

        stat_mock.return_value.st_size = 1000
        stat_mock.return_value.st_uid = 123
        getpwuid_mock.return_value.pw_name = "guest_tmp987"
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])

    def test_unwritable_report(self):
        """get_all_reports() and get_new_reports() for unwritable report"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])

        r = os.path.join(apport.fileutils.report_dir, "unwritable.crash")
        with open(r, "w", encoding="utf-8") as fd:
            fd.write("unwritable")
        os.chmod(r, 0o444)

        if os.getuid() == 0:
            self.assertEqual(set(apport.fileutils.get_new_reports()), set([r]))
            self.assertEqual(set(apport.fileutils.get_all_reports()), set([r]))
        else:
            self.assertEqual(set(apport.fileutils.get_new_reports()), set())
            self.assertEqual(set(apport.fileutils.get_all_reports()), set())

    def test_delete_report(self):
        """delete_report()"""
        tr = self._create_reports()

        while tr:
            self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))
            apport.fileutils.delete_report(tr.pop())

    def test_make_report_file(self):
        """make_report_file()"""
        pr = problem_report.ProblemReport()
        self.assertRaises(ValueError, apport.fileutils.make_report_file, pr)

        pr["Package"] = "bash 1"
        with apport.fileutils.make_report_file(pr) as f:
            path = f.name
            self.assertTrue(
                path.startswith(f"{apport.fileutils.report_dir}/bash"), path
            )
            os.unlink(path)

        pr["ExecutablePath"] = "/bin/bash"
        with apport.fileutils.make_report_file(pr) as f:
            path = f.name
            self.assertTrue(
                path.startswith(f"{apport.fileutils.report_dir}/_bin_bash"), path
            )

        # file exists already, should fail now
        self.assertRaises(OSError, apport.fileutils.make_report_file, pr)

        # should still fail if it's a dangling symlink
        os.unlink(path)
        os.symlink(os.path.join(apport.fileutils.report_dir, "pwned"), path)
        self.assertRaises(OSError, apport.fileutils.make_report_file, pr)

    def test_check_files_md5(self):
        """check_files_md5()"""
        f1 = os.path.join(apport.fileutils.report_dir, "test 1.txt")
        f2 = os.path.join(apport.fileutils.report_dir, "test:2.txt")
        sumfile = os.path.join(apport.fileutils.report_dir, "sums.txt")
        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")
        with open(f2, "w", encoding="utf-8") as fd:
            fd.write("More stuff")
        # use one relative and one absolute path in checksums file
        with open(sumfile, "w", encoding="utf-8") as fd:
            fd.write(
                f"""2e41290da2fa3f68bd3313174467e3b5  {f1[1:]}
f6423dfbc4faf022e58b4d3f5ff71a70  {f2}
"""
            )
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [], "correct md5sums"
        )

        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff!")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [f1[1:]], "file 1 wrong"
        )
        with open(f2, "w", encoding="utf-8") as fd:
            fd.write("More stuff!")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile),
            [f1[1:], f2],
            "files 1 and 2 wrong",
        )
        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [f2], "file 2 wrong"
        )

    def test_get_config(self):
        """get_config()"""
        # nonexisting
        apport.fileutils._config_file = "/nonexisting"
        self.assertIsNone(apport.fileutils.get_config("main", "foo"))
        self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
        apport.fileutils._get_config_parser.cache_clear()

        # empty
        with tempfile.NamedTemporaryFile() as f:
            apport.fileutils._config_file = f.name
            self.assertIsNone(apport.fileutils.get_config("main", "foo"))
            self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
            apport.fileutils._get_config_parser.cache_clear()

            # nonempty
            f.write(
                b"[main]\none=1\ntwo = TWO\nb1 = 1\nb2=False\n" b"[spethial]\none= 99\n"
            )
            f.flush()
            self.assertIsNone(apport.fileutils.get_config("main", "foo"))
            self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
            self.assertEqual(apport.fileutils.get_config("main", "one"), "1")
            self.assertEqual(
                apport.fileutils.get_config("main", "one", default="moo"), "1"
            )
            self.assertEqual(apport.fileutils.get_config("main", "two"), "TWO")
            self.assertEqual(
                apport.fileutils.get_config("main", "b1", boolean=True), True
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b2", boolean=True), False
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b3", boolean=True), None
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b3", default=False, boolean=True),
                False,
            )
            self.assertEqual(apport.fileutils.get_config("spethial", "one"), "99")
            self.assertIsNone(apport.fileutils.get_config("spethial", "two"))
            self.assertEqual(
                apport.fileutils.get_config("spethial", "one", "moo"), "99"
            )
            self.assertEqual(
                apport.fileutils.get_config("spethial", "nope", "moo"), "moo"
            )
            apport.fileutils._get_config_parser.cache_clear()

            # interpolation
            f.write(b"[inter]\none=1\ntwo = TWO\ntest = %(two)s\n")
            f.flush()
            self.assertEqual(apport.fileutils.get_config("inter", "one"), "1")
            self.assertEqual(apport.fileutils.get_config("inter", "two"), "TWO")
            self.assertEqual(apport.fileutils.get_config("inter", "test"), "%(two)s")
            apport.fileutils._get_config_parser.cache_clear()

    def test_shared_libraries(self):
        """shared_libraries()"""
        libs = apport.fileutils.shared_libraries(sys.executable)
        self.assertGreater(len(libs), 3)
        self.assertIn("libc.so.6", libs)
        self.assertIn("libc.so.6", libs["libc.so.6"])
        self.assertTrue(os.path.exists(libs["libc.so.6"]))
        for library_name, library_path in libs.items():
            self.assertNotIn("vdso", library_name, libs)
            self.assertTrue(os.path.exists(library_path))

        self.assertEqual(apport.fileutils.shared_libraries("/non/existing"), {})
        self.assertEqual(apport.fileutils.shared_libraries("/etc"), {})
        self.assertEqual(apport.fileutils.shared_libraries("/etc/passwd"), {})

    def test_links_with_shared_library(self):
        """links_with_shared_library()"""
        self.assertTrue(
            apport.fileutils.links_with_shared_library(sys.executable, "libc")
        )
        self.assertTrue(
            apport.fileutils.links_with_shared_library(sys.executable, "libc.so.6")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library(sys.executable, "libc.so.7")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library(sys.executable, "libd")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library("/non/existing", "libc")
        )
        self.assertFalse(apport.fileutils.links_with_shared_library("/etc", "libc"))
        self.assertFalse(
            apport.fileutils.links_with_shared_library("/etc/passwd", "libc")
        )

    def test_get_core_path(self):
        """get_core_path()"""
        boot_id = apport.fileutils.get_boot_id()

        # Basic test
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test", uid=234, timestamp=222222
        )
        expected = f"core._usr_bin_test.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test dots in exe names
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test.sh", uid=234, timestamp=222222
        )
        expected = f"core._usr_bin_test_sh.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test no exe name
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe=None, uid=234, timestamp=222222
        )
        expected = f"core.unknown.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test no uid
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test", uid=None, timestamp=222222
        )
        expected = f"core._usr_bin_test.{str(os.getuid())}.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

    def test_clean_core_directory(self):
        """clean_core_directory()"""
        fake_uid = 5150
        extra_core_files = 4
        num_core_files = apport.fileutils.max_corefiles_per_uid + extra_core_files

        # Create some test files
        for x in range(num_core_files):
            core_path = apport.fileutils.get_core_path(
                pid=123 + x, exe="/usr/bin/test", uid=fake_uid, timestamp=222222 + x
            )[1]
            with open(core_path, "w", encoding="utf-8") as fd:
                fd.write("Some stuff")
                time.sleep(1)

        # Create a file with a different uid
        core_path = apport.fileutils.get_core_path(
            pid=231, exe="/usr/bin/test", uid=fake_uid + 1, timestamp=333333
        )[1]
        with open(core_path, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")

        # Make sure we have the proper number of test files
        self.assertEqual(
            num_core_files, len(apport.fileutils.find_core_files_by_uid(fake_uid))
        )
        self.assertEqual(1, len(apport.fileutils.find_core_files_by_uid(fake_uid + 1)))

        # Clean out the directory
        apport.fileutils.clean_core_directory(fake_uid)

        # Make sure we have the proper number of test files. We should
        # have one less than max_corefiles_per_uid.
        self.assertEqual(
            apport.fileutils.max_corefiles_per_uid - 1,
            len(apport.fileutils.find_core_files_by_uid(fake_uid)),
        )
        self.assertEqual(1, len(apport.fileutils.find_core_files_by_uid(fake_uid + 1)))

        # Make sure we deleted the oldest ones
        for x in range(apport.fileutils.max_corefiles_per_uid - 1):
            offset = extra_core_files + x + 1
            core_path = apport.fileutils.get_core_path(
                pid=123 + offset,
                exe="/usr/bin/test",
                uid=fake_uid,
                timestamp=222222 + offset,
            )[1]
            self.assertTrue(os.path.exists(core_path))



"""Functions to manage apport problem report files."""

# Copyright (C) 2006 - 2009 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# TODO: Address following pylint complaints
# pylint: disable=invalid-name,missing-docstring

import configparser
import contextlib
import functools
import glob
import http.client
import io
import json
import operator
import os
import pwd
import re
import socket
import stat
import subprocess
import time

from apport.packaging_impl import impl as packaging
from problem_report import ProblemReport

report_dir = os.environ.get("APPORT_REPORT_DIR", "/var/crash")
core_dir = os.environ.get("APPORT_COREDUMP_DIR", "/var/lib/apport/coredump")
max_corefiles_per_uid = 5

_config_file = "~/.config/apport/settings"

SNAPD_SOCKET = "/run/snapd.socket"


#  UHTTPConnection is based on code from the UpdateManager package:
#  Copyright (c) 2017 Canonical
#  Author: Andrea Azzarone <andrea.azzarone@canonical.com>
class UHTTPConnection(http.client.HTTPConnection):
    def __init__(self, path):
        http.client.HTTPConnection.__init__(self, "localhost")
        self.path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self.path)
        self.sock = sock


def allowed_to_report():
    """Check whether crash reporting is enabled."""
    if not os.access("/usr/bin/whoopsie", os.X_OK):
        return True

    try:
        cmd = ["/bin/systemctl", "-q", "is-enabled", "whoopsie.path"]
        return subprocess.call(cmd) == 0
    except OSError:
        return False


def get_dbus_socket(dbus_addr):
    """Extract the socket from a DBus address."""
    if not dbus_addr:
        return None

    # Only support unix domain sockets, and only the default Ubuntu path
    if not dbus_addr.startswith("unix:path=/run/user/"):
        return None

    # Prevent path traversal
    if "../" in dbus_addr:
        return None

    # Don't support escaped values, multiple addresses, or multiple keys
    # and values
    for search in ("%", ",", ";"):
        if search in dbus_addr:
            return None

    parts = dbus_addr.split("=")
    if len(parts) != 2:
        return None

    return parts[1]


def find_package_desktopfile(package):
    """Return a package's .desktop file.

    If given package is installed and has a single .desktop file, return the
    path to it, otherwise return None.
    """
    if package is None:
        return None

    desktopfile = None

    for line in packaging.get_files(package):
        if line.endswith(".desktop"):
            # restrict to autostart and applications, see LP#1147528
            if not line.startswith("/etc/xdg/autostart") and not line.startswith(
                "/usr/share/applications/"
            ):
                continue

            if desktopfile:
                return None  # more than one

            # only consider visible ones
            try:
                with open(line, "rb") as f:
                    if b"NoDisplay=true" not in f.read():
                        desktopfile = line
            except FileNotFoundError:
                continue

    return desktopfile


def likely_packaged(file):
    """Check whether the given file is likely to belong to a package.

    This is semi-decidable: A return value of False is definitive, a True value
    is only a guess which needs to be checked with find_file_package().
    However, this function is very fast and does not access the package
    database.
    """
    # packages only ship executables in these directories
    pkg_allowlist = [
        "/bin/",
        "/boot",
        "/etc/",
        "/initrd",
        "/lib",
        "/sbin/",
        "/opt",
        "/usr/",
        "/var",
    ]

    allowlist_match = False
    for i in pkg_allowlist:
        if file.startswith(i):
            allowlist_match = True
            break
    return (
        allowlist_match
        and not file.startswith("/usr/local/")
        and not file.startswith("/var/lib/")
    )


def find_file_package(file):
    """Return the package that ships the given file.

    Return None if no package ships it.
    """
    # resolve symlinks in directories
    (directory, name) = os.path.split(file)
    resolved_dir = os.path.realpath(directory)
    if os.path.isdir(resolved_dir):
        file = os.path.join(resolved_dir, name)

    if not likely_packaged(file):
        return None

    return packaging.get_file_package(file)


def find_snap(snap):
    """Return the data of the given snap.

    Return None if the snap is not found to be installed.
    """
    try:
        with contextlib.closing(UHTTPConnection(SNAPD_SOCKET)) as c:
            url = f"/v2/snaps/{snap}"
            c.request("GET", url)
            response = c.getresponse()
            if response.status == 200:
                return json.loads(response.read())["result"]
    except (http.client.HTTPException, json.JSONDecodeError, OSError):
        return None
    return None


def seen_report(report):
    """Check whether the report file has already been processed earlier."""
    st = os.stat(report)
    return (st.st_atime > st.st_mtime) or (st.st_size == 0)


def mark_report_upload(report):
    upload = f"{report.rsplit('.', 1)[0]}.upload"
    uploaded = f"{report.rsplit('.', 1)[0]}.uploaded"
    # if uploaded exists and is older than the report remove it and upload
    if os.path.exists(uploaded) and os.path.exists(upload):
        report_st = os.stat(report)
        upload_st = os.stat(upload)
        if upload_st.st_mtime < report_st.st_mtime:
            os.unlink(upload)
    with open(upload, "a", encoding="utf-8"):
        pass


def mark_hanging_process(report, pid):
    if "ExecutablePath" in report:
        subject = report["ExecutablePath"].replace("/", "_")
    else:
        raise ValueError("report does not have the ExecutablePath attribute")

    uid = os.geteuid()
    base = f"{subject}.{str(uid)}.{pid}.hanging"
    path = os.path.join(report_dir, base)
    with open(path, "a", encoding="utf-8"):
        pass


def mark_report_seen(report):
    """Mark given report file as seen."""
    st = os.stat(report)
    try:
        os.utime(report, (st.st_mtime, st.st_mtime - 1))
    except OSError:
        # file is probably not our's, so do it the slow and boring way
        # change the file's access time until it stat's different than the
        # mtime. This might take a while if we only have 1-second resolution.
        # Time out after 1.2 seconds.
        timeout = 12
        while timeout > 0:
            with open(report, encoding="utf-8") as report_file:
                report_file.read(1)
            try:
                st = os.stat(report)
            except OSError:
                return

            if st.st_atime > st.st_mtime:
                break
            time.sleep(0.1)
            timeout -= 1

        if timeout == 0:
            # happens on noatime mounted partitions; just give up and delete
            delete_report(report)


_LOGIN_DEFS_RE = re.compile(
    r"^\s*(?P<name>[A-Z0-9_]+)\s+(?P<quote>[\"\']?)(?P<value>.*)(?P=quote)\s*$"
)


def _parse_login_defs(lines: io.TextIOWrapper) -> dict[str, str]:
    defs = {}
    for line in lines:
        match = _LOGIN_DEFS_RE.match(line)
        if not match:
            continue
        defs[match.group("name")] = match.group("value")
    return defs


@functools.cache
def get_login_defs() -> dict[str, str]:
    """Parse /etc/login.defs and return a dictionary with its content."""
    try:
        with open("/etc/login.defs", encoding="utf-8") as login_defs_file:
            return _parse_login_defs(login_defs_file)
    except FileNotFoundError:
        return {}


def get_sys_gid_max() -> int:
    """Return maximum system group ID (SYS_GID_MAX from /etc/login.defs)."""
    try:
        return int(get_login_defs()["SYS_GID_MAX"])
    except (KeyError, ValueError):
        return 999


def get_sys_uid_max() -> int:
    """Return maximum system user ID (SYS_UID_MAX from /etc/login.defs)."""
    try:
        return int(get_login_defs()["SYS_UID_MAX"])
    except (KeyError, ValueError):
        return 999


def get_all_reports():
    """Return a list with all report files accessible to the calling user."""
    reports = []
    for r in glob.glob(os.path.join(report_dir, "*.crash")):
        try:
            if os.path.getsize(r) > 0 and os.access(r, os.R_OK | os.W_OK):
                reports.append(r)
        except OSError:
            # race condition, can happen if report disappears between glob and
            # stat
            pass
    return reports


def get_new_reports():
    """Get new reports for calling user.

    Return a list with all report files which have not yet been processed
    and are accessible to the calling user.
    """
    reports = []
    for r in get_all_reports():
        try:
            if not seen_report(r):
                reports.append(r)
        except OSError:
            # race condition, can happen if report disappears between glob and
            # stat
            pass
    return reports


def get_all_system_reports():
    """Get all system reports.

    Return a list with all report files which belong to a system user.
    The maximum system user group ID is taken from SYS_UID_MAX from
    /etc/login.defs (defaults to 999 on Debian based systems and LSB
    specifies 499 in "User ID Ranges").
    """
    reports = []
    sys_uid_max = get_sys_uid_max()
    for r in glob.glob(os.path.join(report_dir, "*.crash")):
        try:
            st = os.stat(r)
            if st.st_size > 0 and st.st_uid <= sys_uid_max:
                # filter out guest session crashes;
                # they might have a system UID
                try:
                    pw = pwd.getpwuid(st.st_uid)
                    if pw.pw_name.startswith("guest"):
                        continue
                except KeyError:
                    pass

                reports.append(r)
        except OSError:
            # race condition, can happen if report disappears between glob and
            # stat
            pass
    return reports


def get_new_system_reports():
    """Get new system reports.

    Return a list with all report files which have not yet been processed
    and belong to a system user. The maximum system user group ID is taken
    from SYS_UID_MAX from /etc/login.defs (defaults to 999 on Debian based
    systems and LSB specifies 499 in "User ID Ranges").
    """
    return [r for r in get_all_system_reports() if not seen_report(r)]


def delete_report(report):
    """Delete the given report file.

    If unlinking the file fails due to a permission error (if report_dir is not
    writable to normal users), the file will be truncated to 0 bytes instead.
    """
    try:
        os.unlink(report)
    except OSError:
        with open(report, "w", encoding="utf-8") as f:
            f.truncate(0)


def get_recent_crashes(report):
    """Return the number of recent crashes for the given report file.

    Return the number of recent crashes (currently, crashes which happened more
    than 24 hours ago are discarded).
    """
    pr = ProblemReport()
    pr.load(report, False, key_filter=["CrashCounter", "Date"])
    try:
        count = int(pr["CrashCounter"])
        report_time = time.mktime(time.strptime(pr["Date"]))
        cur_time = time.mktime(time.localtime())
        # discard reports which are older than 24 hours
        if cur_time - report_time > 24 * 3600:
            return 0
        return count
    except (ValueError, KeyError):
        return 0


def increment_crash_counter(report: ProblemReport, filename: str) -> None:
    """Increment the crash counter if report was seen."""
    if not seen_report(filename):
        return
    # Make sure the file isn't a FIFO or symlink
    fd = os.open(filename, os.O_NOFOLLOW | os.O_RDONLY | os.O_NONBLOCK)
    st = os.fstat(fd)
    if stat.S_ISREG(st.st_mode):
        with os.fdopen(fd, "rb") as f:
            crash_counter = get_recent_crashes(f) + 1
        report["CrashCounter"] = str(crash_counter)


def make_report_file(report, uid=None):
    """Construct a canonical pathname for a report and open it for writing.

    If uid is not given, it defaults to the effective uid of the current
    process. The report file must not exist already, to prevent losing
    previous reports or symlink attacks.

    Return an open file object for binary writing.
    """
    if "ExecutablePath" in report:
        subject = report["ExecutablePath"].replace("/", "_")
    elif "Package" in report:
        subject = report["Package"].split(None, 1)[0]
    else:
        raise ValueError("report has neither ExecutablePath nor Package attribute")

    if not uid:
        uid = os.geteuid()

    path = os.path.join(report_dir, f"{subject}.{str(uid)}.crash")
    return open(path, "xb")


def check_files_md5(sumfile):
    """Check file integrity against md5 sum file.

    sumfile must be md5sum(1) format (relative to /).

    Return a list of files that don't match.
    """
    assert os.path.exists(sumfile)
    md5sum = subprocess.run(
        ["/usr/bin/md5sum", "-c", sumfile],
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd="/",
        env={},
    )

    # if md5sum succeeded, don't bother parsing the output
    if md5sum.returncode == 0:
        return []

    mismatches = []
    for line in md5sum.stdout.decode().splitlines():
        if line.endswith("FAILED"):
            mismatches.append(line.rsplit(":", 1)[0])

    return mismatches


@functools.cache
def _get_config_parser(path: str) -> configparser.ConfigParser:
    config = configparser.ConfigParser(interpolation=None)
    contents = ""
    fd = None
    f = None
    try:
        fd = os.open(path, os.O_NOFOLLOW | os.O_RDONLY)
        st = os.fstat(fd)
        if stat.S_ISREG(st.st_mode):
            f = os.fdopen(fd, "r")
            # Limit size to prevent DoS
            contents = f.read(500)
    except OSError:
        pass
    finally:
        if f is not None:
            f.close()
        elif fd is not None:
            os.close(fd)

    try:
        config.read_string(contents)
    except configparser.MissingSectionHeaderError:
        pass

    return config


def get_config(section, setting, default=None, path=None, boolean=False):
    """Return a setting from user configuration.

    This is read from ~/.config/apport/settings or path. If bool is True, the
    value is interpreted as a boolean.

    Privileges may need to be dropped before calling this.
    """
    if not path:
        # Properly handle dropped privileges
        homedir = pwd.getpwuid(os.geteuid())[5]
        path = _config_file.replace("~", homedir)

    config = _get_config_parser(path)

    try:
        if boolean:
            return config.getboolean(section, setting)
        return config.get(section, setting)
    except (configparser.NoOptionError, configparser.NoSectionError):
        return default


def get_starttime(contents: str) -> int:
    """Extract the starttime from the contents of a stat file."""
    # 22nd field in a stat file is the time the process started after
    # system boot in clock ticks. In order to prevent filename
    # manipulations including spaces or extra parentheses, skip all the way
    # to the very last closing parentheses, then start counting.
    stripped = contents[contents.rfind(")") + 2 :]
    # We've skipped over the PID and the filename, so index is now 19.
    return int(stripped.split()[19])


def get_uid_and_gid(contents):
    """Extract the uid and gid from the contents of a status file."""
    real_uid = None
    real_gid = None
    for line in contents.splitlines():
        # Iterate through the whole contents to make sure we're getting
        # the last Uid and Gid lines in the file and not a manipulated
        # process name with embedded newlines.
        if line.startswith("Uid:") and len(line.split()) > 1:
            real_uid = int(line.split()[1])
        elif line.startswith("Gid:") and len(line.split()) > 1:
            real_gid = int(line.split()[1])
    return (real_uid, real_gid)


def search_map(mapfd, uid):
    """Search for an ID in a map fd."""
    for line in mapfd:
        fields = line.split()
        if len(fields) != 3:
            continue

        host_start = int(fields[1])
        host_end = host_start + int(fields[2])

        if host_start <= uid <= host_end:
            return True

    return False


def get_boot_id():
    """Get the kernel boot id."""
    with open("/proc/sys/kernel/random/boot_id", encoding="utf-8") as f:
        boot_id = f.read().strip()
    return boot_id


def get_process_environ(proc_pid_fd: int) -> dict[str, str]:
    """Get the process environ from a proc directory file descriptor.

    Raises an OSError in case the environ file could not been read.
    """

    def opener(path: str | os.PathLike[str], flags: int) -> int:
        return os.open(path, flags, dir_fd=proc_pid_fd)

    with open(
        "environ", encoding="utf-8", errors="replace", opener=opener
    ) as environ_fd:
        environ = environ_fd.read().rstrip("\0 ")

    if not environ:
        return {}
    return dict([entry.split("=", 1) for entry in environ.split("\0") if "=" in entry])


def get_process_path(proc_pid_fd=None):
    """Get the process path from a proc directory file descriptor."""
    if proc_pid_fd is None:
        return "unknown"
    try:
        return os.readlink("exe", dir_fd=proc_pid_fd)
    except OSError:
        return "unknown"


def get_core_path(pid=None, exe=None, uid=None, timestamp=None, proc_pid_fd=None):
    """Get the path to a core file."""

    def opener(path: str | os.PathLike[str], flags: int) -> int:
        return os.open(path, flags, dir_fd=proc_pid_fd)

    if pid is None:
        pid = "unknown"
        timestamp = "unknown"
    elif timestamp is None:
        if proc_pid_fd is not None:
            stat_filename = "stat"
        else:
            stat_filename = f"/proc/{pid}/stat"
        with open(stat_filename, encoding="utf-8", opener=opener) as stat_file:
            stat_contents = stat_file.read()
        timestamp = get_starttime(stat_contents)

    if exe is None:
        exe = get_process_path(proc_pid_fd)
    exe = exe.replace("/", "_").replace(".", "_")

    if uid is None:
        uid = os.getuid()

    # This is similar to systemd-coredump, but with the exe name instead
    # of the command name
    core_name = f"core.{exe}.{uid}.{get_boot_id()}.{str(pid)}.{str(timestamp)}"

    core_path = os.path.join(core_dir, core_name)

    return (core_name, core_path)


def find_core_files_by_uid(uid):
    """Search the core file directory for files that belong to a
    specified uid. Returns a list of lists containing the filename and
    the file modification time.
    """
    uid = str(uid)
    core_files = []
    uid_files = []

    if os.path.exists(core_dir):
        core_files = os.listdir(path=core_dir)

    for f in core_files:
        try:
            if f.split(".")[2] == uid:
                core_file_time = os.path.getmtime(os.path.join(core_dir, f))
                uid_files.append([f, core_file_time])
        except (IndexError, FileNotFoundError):
            continue
    return uid_files


def clean_core_directory(uid):
    """Remove old files from the core directory if there are more than
    the maximum allowed per uid.
    """
    uid_files = find_core_files_by_uid(uid)
    sorted_files = sorted(uid_files, key=operator.itemgetter(1))

    # Subtract a extra one to make room for the new core file
    if len(uid_files) > max_corefiles_per_uid - 1:
        for _ in range(len(uid_files) - max_corefiles_per_uid + 1):
            os.remove(os.path.join(core_dir, sorted_files[0][0]))
            sorted_files.remove(sorted_files[0])


def shared_libraries(path):
    """Get libraries with which the specified binary is linked.

    Return a library name -> path mapping, for example 'libc.so.6' ->
    '/lib/x86_64-linux-gnu/libc.so.6'.
    """
    libs = {}

    with subprocess.Popen(
        ["ldd", path],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    ) as ldd:
        for line in ldd.stdout:
            try:
                name, rest = line.split("=>", 1)
            except ValueError:
                continue

            name = name.strip()
            # exclude linux-vdso since that is a virtual so
            if "linux-vdso" in name:
                continue
            # this is usually "path (address)"
            rest = rest.split()[0].strip()
            if rest.startswith("("):
                continue
            libs[name] = rest
        ldd.stdout.close()

    if ldd.returncode != 0:
        return {}
    return libs


def should_skip_crash(report: ProblemReport, filename: str) -> str | None:
    """Check if the crash should be skipped for flood protection.

    In case the crash should be skipped return a string with the reason.
    Otherwise return None.
    """
    try:
        crash_counter = int(report["CrashCounter"])
    except (KeyError, ValueError):
        crash_counter = 0
    if crash_counter > 1:
        return f"this executable already crashed {crash_counter} times, ignoring"
    if os.path.exists(filename) and not seen_report(filename):
        # don't clobber existing report
        return (
            f"report {filename} already exists and unseen,"
            f" skipping to avoid disk usage DoS"
        )
    return None


def links_with_shared_library(path, lib):
    """Check if the binary at path links with the library named lib.

    path should be a fully qualified path (e.g. report['ExecutablePath']),
    lib may be of the form 'lib<name>' or 'lib<name>.so.<version>'
    """
    libs = shared_libraries(path)

    if lib in libs:
        return True

    for linked_lib in libs:
        if linked_lib.startswith(f"{lib}.so."):
            return True

    return False



"""Attach generally useful information, not specific to any package."""

# Copyright (C) 2009 Canonical Ltd.
# Authors: Matt Zimmerman <mdz@canonical.com>
#          Martin Pitt <martin.pitt@ubuntu.com>
#          Brian Murray <brian@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import os
import re

import apport.fileutils
import apport.hookutils
import apport.ui
from problem_report import ProblemReport


def add_info(report: ProblemReport, ui: apport.ui.HookUI) -> None:
    """Add generic, general information to the problem report."""
    # TODO: Split into smaller functions/methods
    # pylint: disable=too-many-branches,too-many-locals
    nonfree_modules = apport.hookutils.nonfree_kernel_modules()
    if nonfree_modules:
        report["NonfreeKernelModules"] = " ".join(nonfree_modules)

    # check for low space
    mounts = {"/": "system", "/var": "/var", "/tmp": "/tmp"}

    home = os.getenv("HOME")
    if home:
        mounts[home] = "home"
    threshold = 50

    for mount, mount_name in mounts.items():
        try:
            stat = os.statvfs(mount)
        except FileNotFoundError:
            continue
        free_mb = stat.f_bavail * stat.f_frsize / 1000000

        if free_mb < threshold:
            report["UnreportableReason"] = (
                f"Your {mount_name} partition has less than {free_mb} MB"
                f" of free space available, which leads to problems using"
                f" applications and installing updates."
                f" Please free some space."
            )

    # important glib errors/assertions (which should not have private data)
    if "ExecutablePath" in report:
        path = report["ExecutablePath"]
        gtk_like = (
            apport.fileutils.links_with_shared_library(path, "libgtk")
            or apport.fileutils.links_with_shared_library(path, "libgtk-3")
            or apport.fileutils.links_with_shared_library(path, "libX11")
        )
        if gtk_like and apport.hookutils.in_session_of_problem(report):
            xsession_errors = apport.hookutils.xsession_errors()
            if xsession_errors:
                report["XsessionErrors"] = xsession_errors

    # using local libraries?
    if "ProcMaps" in report:
        local_libs = set()
        for lib in re.finditer(r"\s(/[^ ]+\.so[.0-9]*)$", report["ProcMaps"], re.M):
            if not apport.fileutils.likely_packaged(lib.group(1)):
                local_libs.add(lib.group(1))
        if ui and local_libs:
            if not ui.yesno(
                """\
The crashed program seems to use third-party or local libraries:

%s

It is highly recommended to check if the problem persists without those first.

Do you want to continue the report process anyway?
"""
                % "\n".join(local_libs)
            ):
                raise StopIteration
            report["LocalLibraries"] = " ".join(local_libs)
            report.add_tags(["local-libs"])

    # using third-party packages?
    if "[origin:" in report.get("Package", "") or "[origin:" in report.get(
        "Dependencies", ""
    ):
        report.add_tags(["third-party-packages"])

    # using ecryptfs?
    if os.path.exists(os.path.expanduser("~/.ecryptfs/wrapped-passphrase")):
        report["EcryptfsInUse"] = "Yes"

    # filter out crashes on missing GLX (LP#327673)
    in_gl = "/usr/lib/libGL.so" in (report.get("StacktraceTop") or "\n").splitlines()[0]
    if in_gl and "Loading extension GLX" not in apport.hookutils.read_file(
        "/var/log/Xorg.0.log"
    ):
        report["UnreportableReason"] = (
            "The X.org server does not support the GLX extension,"
            " which the crashed program expected to use."
        )
    # filter out package install failures due to a segfault
    if (
        "Segmentation fault" in report.get("ErrorMessage", "")
        and report["ProblemType"] == "Package"
    ):
        report["UnreportableReason"] = (
            "The package installation resulted in a segmentation fault which"
            " is better reported as a crash report rather than a package"
            " install failure."
        )

    # log errors
    if report["ProblemType"] == "Crash":
        apport.hookutils.attach_journal_errors(report)


if __name__ == "__main__":
    r = ProblemReport()
    add_info(r, apport.ui.NoninteractiveHookUI())
    for key, value in r.items():
        print(f"{key}: {value}")



#!/usr/bin/python3
#
# Copyright 2009-2010  Canonical, Ltd.
# Author: Kees Cook <kees@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Examine the crash files saved by apport to attempt to determine the cause
of a segfault.  Currently very very simplistic, and only finds commonly
understood situations for x86/x86_64."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name,missing-class-docstring,missing-function-docstring

import logging
import re
import sys


class ParseSegv:
    def __init__(self, registers, disassembly, maps):
        self.regs = self.parse_regs(registers)
        self.sp = None
        for reg in ("rsp", "esp"):
            if reg in self.regs:
                self.sp = self.regs[reg]

        (self.line, self.pc, self.insn, self.src, self.dest) = self.parse_disassembly(
            disassembly
        )

        self.stack_vma = None
        self.maps = self.parse_maps(maps)

    def find_vma(self, addr):
        for vma in self.maps:
            if vma["start"] <= addr < vma["end"]:
                return vma
        return None

    def parse_maps(self, maps_str):
        maps = []
        for line in maps_str.splitlines():
            items = line.strip().split()
            if len(items) < 4:
                raise ValueError(f"Cannot parse maps line: {line.strip()}")
            span, perms = items[0:2]
            if len(items) == 5:
                name = None
            else:
                name = items[5]
            start, end = [int(x, 16) for x in span.split("-")]
            if name == "[stack]":
                self.stack_vma = len(maps)
            maps.append({"start": start, "end": end, "perms": perms, "name": name})
            logging.debug(
                "start: %s, end: %s, perms: %s, name: %s", start, end, perms, name
            )
        return maps

    @staticmethod
    def parse_regs(reg_str: str) -> dict[str, int]:
        regs = {}
        for line in reg_str.splitlines():
            reg, hexvalue = line.split()[0:2]
            if hexvalue == "<unavailable>":
                continue
            regs[reg] = int(hexvalue, 16)
            logging.debug("%s:0x%08x", reg, regs[reg])
        return regs

    def parse_disassembly(self, disassembly):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches
        if not self.regs:
            raise ValueError("Registers not loaded yet!?")
        lines = disassembly.splitlines()
        # Throw away possible 'Dump' gdb report line
        if len(lines) > 0 and lines[0].startswith("Dump"):
            lines.pop(0)
        if len(lines) < 1:
            raise ValueError("Failed to load empty disassembly")
        line = lines[0].strip()
        # Drop GDB 7.1's leading $pc mark
        if line.startswith("=>"):
            line = line[2:].strip()
        logging.debug(line)
        pc_str = line.split()[0]
        if pc_str.startswith("0x"):
            pc = int(pc_str.split(":")[0], 16)
        else:
            # Could not identify this instruction line
            raise ValueError(
                f'Could not parse PC "{pc_str}" from disassembly line: {line}'
            )
        logging.debug("pc: 0x%08x", pc)

        full_insn_str = line.split(":", 1)[1].strip()
        # Handle invalid memory
        if "Cannot access memory at address" in full_insn_str or (
            full_insn_str == "" and len(lines) == 1
        ):
            return line, pc, None, None, None
        # Handle wrapped lines
        if full_insn_str == "" and lines[1].startswith(" "):
            line = f"{line} {lines[1].strip()}"
            full_insn_str = line.split(":", 1)[1].strip()

        insn_parts = full_insn_str.split()
        # Drop call target names "call   0xb7a805af <_Unwind_Find_FDE@plt+111>"
        if insn_parts[-1].endswith(">") and insn_parts[-1].startswith("<"):
            insn_parts.pop(-1)
        # Attempt to find arguments
        args_str = ""
        if len(insn_parts) > 1:
            args_str = insn_parts.pop(-1)
        # Assume remainder is the insn itself
        insn = " ".join(insn_parts)
        logging.debug("insn: %s", insn)

        args = []
        src = None
        dest = None
        if args_str == "":
            # Could not find insn args
            args = None
        else:
            logging.debug('args: "%s"', args_str)

            for m in re.finditer(r"([^,\(]*(\(:?[^\)]+\))*)", args_str):
                if len(m.group(0)):
                    args.append(m.group(0))
            if len(args) > 0:
                src = args[0]
                logging.debug("src: %s", src)
            if len(args) > 1:
                dest = args[1]
                logging.debug("dest: %s", dest)

        # Set up possible implicit memory destinations (stack actions)
        if insn in {"push", "pop", "pushl", "popl", "call", "callq", "ret", "retq"}:
            for reg in ("rsp", "esp"):
                if reg in self.regs:
                    dest = f"(%{reg})"
                    break

        return line, pc, insn, src, dest

    def validate_vma(self, perm, addr, name):
        perm_name = {
            "x": ["executable", "executing"],
            "r": ["readable", "reading"],
            "w": ["writable", "writing"],
        }
        vma = self.find_vma(addr)
        if vma is None:
            alarmist = "unknown"
            if addr < 65536:
                alarmist = "NULL"
            return (
                False,
                f"{name} (0x{addr:08x}) not located in a known VMA region"
                f" (needed {perm_name[perm][0]} region)!",
                f"{perm_name[perm][1]} {alarmist} VMA",
            )
        if perm not in vma["perms"]:
            alarmist = ""
            if perm == "x":
                if "w" in vma["perms"]:
                    alarmist = "writable "
                else:
                    alarmist = "non-writable "
            short = f"{perm_name[perm][1]} {alarmist}VMA {vma['name']}"

            return (
                False,
                f"{name} (0x{addr:08x}) in non-{perm_name[perm][0]} VMA"
                f" region: 0x{vma['start']:08x}-0x{vma['end']:08x}"
                f" {vma['perms']} {vma['name']}",
                short,
            )

        return (True, f"{name} (0x{addr:08x}) ok", f"{perm_name[perm][1]} ok")

    def register_value(self, reg):
        reg_orig = reg

        mask = 0
        if reg.startswith("%"):
            reg = reg[1:]
        if reg in self.regs:
            return self.regs[reg]

        if len(reg) == 2 and reg.endswith("l"):
            mask |= 0xFF00
            reg = f"{reg[0]}x"
        if reg in self.regs:
            return self.regs[reg] & ~mask

        if len(reg) == 2 and reg.endswith("x"):
            mask |= 0xFFFF0000
            reg = f"e{reg}"
        if reg in self.regs:
            return self.regs[reg] & ~mask

        if len(reg) == 3 and reg.startswith("e"):
            mask |= 0xFFFFFFFF00000000
            reg = f"r{reg[1:]}"
        if reg in self.regs:
            return self.regs[reg] & ~mask
        raise ValueError(f"Could not resolve register '{reg_orig}'")

    def calculate_arg(self, arg):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches

        # Check for and pre-remove segment offset
        segment = 0
        if arg.startswith("%") and ":" in arg:
            parts = arg.split(":", 1)
            segment = self.regs[parts[0][1:]]
            arg = parts[1]

        # Handle standard offsets
        parts = arg.split("(")
        offset = parts[0]
        # Handle negative signs
        sign = 1
        if offset.startswith("-"):
            sign = -1
            offset = offset[1:]
        # Skip call target dereferences
        if offset.startswith("*"):
            offset = offset[1:]
        if len(offset) > 0:
            if offset.startswith("%"):
                # Handle the *%REG case
                add = self.regs[offset[1:]]
            else:
                if not offset.startswith("0x"):
                    raise ValueError(f"Unknown offset literal: {parts[0]}")
                add = int(offset[2:], 16) * sign
        else:
            add = 0

        def _reg_val(self, text, val=0):
            if text.startswith("%"):
                val = self.regs[text[1:]]
            elif text == "":
                val = 0
            else:
                val = int(text)
            return val

        # (%ebx, %ecx, 4) style
        value = 0
        if len(parts) > 1:
            parens = parts[1][0:-1]
            reg_list = parens.split(",")

            base = 0
            if len(reg_list) > 0:
                base = _reg_val(self, reg_list[0], base)
            index = 0
            if len(reg_list) > 1:
                index = _reg_val(self, reg_list[1], index)
            scale = 1
            if len(reg_list) > 2:
                scale = _reg_val(self, reg_list[2], scale)
            value = base + index * scale

        value = segment + value + add
        if "esp" in self.regs:
            # 32bit
            return value % 0x100000000
        # 64bit
        return value % 0x10000000000000000

    def report(self):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-statements
        understood = False
        reason = []
        details = [f"Segfault happened at: {self.line}"]

        # Verify PC is in an executable region
        valid, out, short = self.validate_vma("x", self.pc, "PC")
        details.append(out)
        if not valid:
            reason.append(short)
            understood = True

        if self.insn in {"lea", "leal"}:
            # Short-circuit for instructions that do not cause vma access
            details.append(f"insn ({self.insn}) does not access VMA")
        else:
            # Verify source is readable
            if self.src:
                if (
                    ":" not in self.src
                    and (self.src[0] in {"%", "$", "*"})
                    and not self.src.startswith("*%")
                ):
                    details.append(f'source "{self.src}" ok')
                else:
                    addr = self.calculate_arg(self.src)
                    valid, out, short = self.validate_vma(
                        "r", addr, f'source "{self.src}"'
                    )
                    details.append(out)
                    if not valid:
                        reason.append(short)
                        understood = True

            # Verify destination is writable
            if self.dest:
                if ":" not in self.dest and (self.dest[0] in {"%", "$", "*"}):
                    details.append(f'destination "{self.dest}" ok')
                else:
                    addr = self.calculate_arg(self.dest)
                    valid, out, short = self.validate_vma(
                        "w", addr, f'destination "{self.dest}"'
                    )
                    details.append(out)
                    if not valid:
                        reason.append(short)
                        understood = True

        # Handle I/O port operations
        if self.insn in {"out", "in"} and not understood:
            msg = (
                f"disallowed I/O port operation"
                f" on port {self.register_value(self.src)}"
            )
            reason.append(msg)
            details.append(msg)
            understood = True

        # Note position of SP with regard to "[stack]" VMA
        if self.sp is not None:
            if self.stack_vma is not None:
                if self.sp < self.maps[self.stack_vma]["start"]:
                    details.append("Stack memory exhausted (SP below stack segment)")
                if self.sp >= self.maps[self.stack_vma]["end"]:
                    details.append("Stack pointer not within stack segment")
            if not understood:
                valid, out, short = self.validate_vma("r", self.sp, "SP")
                details.append(out)
                if not valid:
                    reason.append(short)
                    understood = True

        if not understood:
            vma = self.find_vma(self.pc)
            msg = "Reason could not be automatically determined."
            if vma and (vma["name"] == "[vdso]" or vma["name"] == "[vsyscall]"):
                msg += " (Unhandled exception in kernel code?)"
            reason.append(msg)
            details.append(msg)
        return understood, "\n".join(reason), "\n".join(details)


def add_info(report):
    # Only interested in segmentation faults...
    if report.get("Signal", "0") != "11":
        return

    needed = ["Signal", "Architecture", "Disassembly", "ProcMaps", "Registers"]
    for field in needed:
        if field not in report:
            report["SegvAnalysis"] = f'Skipped: missing required field "{field}"'
            return

    # Only run on segv for x86 and x86_64...
    if not report["Architecture"] in {"i386", "amd64"}:
        return

    try:
        segv = ParseSegv(report["Registers"], report["Disassembly"], report["ProcMaps"])
        understood, reason, details = segv.report()
        if understood:
            report["SegvReason"] = reason
        report["SegvAnalysis"] = details
    except Exception as error:  # pylint: disable=broad-except
        report["SegvAnalysis"] = f"Failure: {str(error)}"


# pylint: disable-next=missing-function-docstring
def main():
    if len(sys.argv) != 4 or sys.argv[1] in {"-h", "--help"}:
        print("To run self-test, run without any arguments (or with -v)")
        print("To do stand-alone crash parsing:")
        print(f"  Usage: {sys.argv[0]} Registers.txt Disassembly.txt ProcMaps.txt")
        sys.exit(0)

    with open(sys.argv[1], encoding="utf-8") as registers_file:
        registers = registers_file.read()
    with open(sys.argv[2], encoding="utf-8") as disassembly_file:
        disassembly = disassembly_file.read()
    with open(sys.argv[3], encoding="utf-8") as maps_file:
        maps = maps_file.read()
    segv = ParseSegv(registers, disassembly, maps)
    understood, reason, details = segv.report()
    print(f"{reason}\n\n{details}")
    rc = 0
    if not understood:
        rc = 1
    sys.exit(rc)


if __name__ == "__main__":
    main()



"""Debian crash database interface."""

# Debian adaptation Copyright (C) 2012 Ritesh Raj Sarraf <rrs@debian.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.


import email.mime.text
import smtplib
import tempfile

import apport.crashdb


class CrashDatabase(apport.crashdb.CrashDatabase):
    """Debian crash database.

    This is a Apport CrashDB implementation for interacting with Debian BTS.
    """

    # TODO: Implement several missing abstract methods from parent class
    # pylint: disable=abstract-method

    def __init__(self, auth_file, options):
        """Initialize crash database connection.

        Debian implementation is pretty basic as most of its bug management
        processes revolve around the email interface
        """
        apport.crashdb.CrashDatabase.__init__(self, auth_file, options)
        self.options = options

        if not self.options.get("smtphost"):
            self.options["smtphost"] = "reportbug.debian.org"

        if not self.options.get("recipient"):
            self.options["recipient"] = "submit@bugs.debian.org"

    def accepts(self, report):
        """Check if this report can be uploaded to this database.
        Checks for the proper settings of apport.
        """
        if not self.options.get("sender") and "UnreportableReason" not in report:
            report["UnreportableReason"] = (
                "Please configure sender settings in /etc/apport/crashdb.conf"
            )

        # At this time, we are not ready to take CrashDumps
        if "Stacktrace" in report and not report.has_useful_stacktrace():
            report["UnreportableReason"] = (
                "Incomplete backtrace. Please install the debug symbol packages"
            )

        return apport.crashdb.CrashDatabase.accepts(self, report)

    def upload(self, report, progress_callback=None, user_message_callback=None):
        """Upload given problem report return a handle for it.

        In Debian, we use BTS, which is heavily email oriented. This method
        crafts the bug into an email report understood by Debian BTS.
        """
        # first and foremost, let's check if the apport bug filing
        # settings are set correct
        assert self.accepts(report)

        # Frame the report in the format the BTS understands
        try:
            (buggy_package, buggy_version) = report["Package"].split(" ")
        except (KeyError, ValueError):
            return False

        with tempfile.NamedTemporaryFile() as temp:
            temp.file.write(f"Package: {buggy_package}\n".encode("UTF-8"))
            temp.file.write(f"Version: {buggy_version}\n\n\n".encode("UTF-8"))
            temp.file.write(("=============================\n\n").encode("UTF-8"))

            # Let's remove the CoreDump first

            # Even if we have a valid backtrace, we already are reporting it
            # as text. We don't want to send very large emails to the BTS.
            # OTOH, if the backtrace is invalid, has_useful_backtrace() will
            # already deny reporting of the bug report.
            try:
                del report["CoreDump"]
            except KeyError:
                pass

            # Now write the apport bug report
            report.write(temp)

            temp.file.seek(0)

            msg = email.mime.text.MIMEText(temp.file.read().decode("UTF-8"))

        msg["Subject"] = report["Title"]
        msg["From"] = self.options["sender"]
        msg["To"] = self.options["recipient"]

        # Subscribe the submitted to the bug report
        msg.add_header("X-Debbugs-CC", self.options["sender"])
        msg.add_header("Usertag", f"apport-{report['ProblemType'].lower()}")

        smtp = smtplib.SMTP(self.options["smtphost"])
        smtp.sendmail(
            self.options["sender"],
            self.options["recipient"],
            msg.as_string().encode("UTF-8"),
        )
        smtp.quit()
        return True

    def get_comment_url(self, report, handle):
        """Return an URL that should be opened after report has been uploaded
        and upload() returned handle.

        Should return None if no URL should be opened (anonymous filing without
        user comments); in that case this function should do whichever
        interactive steps it wants to perform.
        """
        return None
